from unittest.mock import patch

import numpy as np
import torch
from accelerate import Accelerator

from agilerl.utils.log_utils import DistributeCombineLogs, label_logs


def test_init_DCL():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)

    assert DCL.totals == {}
    assert DCL.accelerator == accelerator
    assert DCL.use_wandb == use_wandb


def test_convert_key():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)

    k = (1,)

    conv = DCL.convert_key(k)

    assert conv == ("__count__", 1)
    assert DCL.key_is_count(conv)


def test_log():
    accelerator = Accelerator()
    use_wandb = True

    with patch("agilerl.training.train_off_policy.wandb.log") as mock_wandb_log:

        def dummy_func(a):
            return a

        DCL = DistributeCombineLogs(accelerator, use_wandb)
        _ = DCL.log(dummy_func)

        mock_wandb_log.assert_called()


def test_accum_logs():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)

    DCL.totals = {("__count__", "a"): [0, 1], ("a",): [1, 2]}
    logs = {"a": [1, 2], "b": [2, 3]}
    DCL.accum_logs(logs)

    assert DCL.totals != {("__count__", "a"): [0, 1], ("a",): [1, 2]}


def test_gather_logs():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)

    DCL.totals = {("__count__", "a"): torch.tensor([0]), ("a",): torch.tensor([1, 2])}

    def dummy_func(a):
        return a

    logs = DCL.gather_logs(dummy_func)

    assert logs == {"a": np.inf}


def test_gather_logs_not_count():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)

    DCL.totals = {
        ("__count__", "a"): torch.tensor([0, 1]),
        ("a",): torch.tensor([1, 2]),
    }

    def dummy_func(a):
        return a

    logs = DCL.gather_logs(dummy_func)

    assert logs == {"a": 3.0}


def test_reset_totals():
    accelerator = Accelerator()
    use_wandb = False

    DCL = DistributeCombineLogs(accelerator, use_wandb)
    DCL.totals = {"asdfghjkl"}
    DCL.reset_logs()

    assert DCL.totals == {}


def test_label_logs():
    logs = "log"
    label = "label"

    labelled = label_logs(logs, label)

    assert labelled == {label: logs}



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer

from agilerl.algorithms.dqn import DQN
from agilerl.components.data import Transition
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import (
    assert_state_dicts_equal,
    get_experiences_batch,
    get_sample_from_space,
)

# Cleanup fixture moved to conftest.py for better performance


class DummyDQN(DQN):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, observation_space, vect=True, num_envs=2):
        self.observation_space = observation_space.shape
        self.vect = vect
        if self.vect:
            self.observation_space = (num_envs,) + self.observation_space
            self.n_envs = num_envs
            self.num_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.observation_space), {}

    def step(self, action):
        return (
            np.random.rand(*self.observation_space),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


# initialize DQN with valid parameters
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
        ("multidiscrete_space", EvolvableMLP),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_dqn(
    observation_space, encoder_cls, accelerator, discrete_space, request
):
    action_space = discrete_space
    observation_space = request.getfixturevalue(observation_space)
    dqn = DQN(observation_space, action_space, accelerator=accelerator)

    expected_device = accelerator.device if accelerator else "cpu"
    assert dqn.observation_space == observation_space
    assert dqn.action_space == action_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == expected_device
    assert dqn.accelerator == accelerator
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]
    assert dqn.double is False
    # assert dqn.actor_network is None
    assert isinstance(dqn.actor.encoder, encoder_cls)
    assert isinstance(dqn.actor_target.encoder, encoder_cls)
    expected_opt_cls = AcceleratedOptimizer if accelerator else optim.Adam
    assert isinstance(dqn.optimizer.optimizer, expected_opt_cls)
    assert isinstance(dqn.criterion, nn.MSELoss)
    dqn.clean_up()


# Can initialize DQN with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        ("vector_space", "simple_mlp", torch.randn(1, 4)),
        (
            "image_space",
            "simple_cnn",
            torch.randn(1, 3, 32, 32),
        ),
    ],
)
def test_initialize_dqn_with_actor_network_make_evo(
    observation_space, actor_network, input_tensor, request, discrete_space
):
    action_space = discrete_space
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    dqn = DQN(observation_space, action_space, actor_network=actor_network)

    assert dqn.observation_space == observation_space
    assert dqn.action_space == action_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == "cpu"
    assert dqn.accelerator is None
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]
    assert dqn.double is False
    assert isinstance(dqn.optimizer.optimizer, optim.Adam)
    assert isinstance(dqn.criterion, nn.MSELoss)
    dqn.clean_up()


@pytest.mark.parametrize(
    "observation_space, net_type",
    [
        ("vector_space", "mlp"),
        ("image_space", "cnn"),
    ],
)
def test_initialize_dqn_with_actor_network_evo_net(
    observation_space, net_type, discrete_space, request
):
    action_space = discrete_space
    observation_space = request.getfixturevalue(observation_space)
    if net_type == "mlp":
        actor_network = EvolvableMLP(
            num_inputs=observation_space.shape[0],
            num_outputs=action_space.n,
            hidden_size=[64, 64],
            activation="ReLU",
        )
    else:
        actor_network = EvolvableCNN(
            input_shape=observation_space.shape,
            num_outputs=action_space.n,
            channel_size=[8, 8],
            kernel_size=[2, 2],
            stride_size=[1, 1],
            activation="ReLU",
        )

    dqn = DQN(observation_space, action_space, actor_network=actor_network)

    assert dqn.observation_space == observation_space
    assert dqn.action_space == action_space
    assert dqn.batch_size == 64
    assert dqn.lr == 0.0001
    assert dqn.learn_step == 5
    assert dqn.gamma == 0.99
    assert dqn.tau == 0.001
    assert dqn.mut is None
    assert dqn.device == "cpu"
    assert dqn.accelerator is None
    assert dqn.index == 0
    assert dqn.scores == []
    assert dqn.fitness == []
    assert dqn.steps == [0]
    assert dqn.double is False
    assert isinstance(dqn.optimizer.optimizer, optim.Adam)
    assert isinstance(dqn.criterion, nn.MSELoss)
    dqn.clean_up()


def test_initialize_dqn_with_incorrect_actor_net_type(vector_space, discrete_space):
    actor_network = "dummy"

    with pytest.raises(TypeError) as a:
        dqn = DQN(vector_space, discrete_space, actor_network=actor_network)

        assert dqn
        assert (
            str(a.value)
            == f"'actor_network' argument is of type {type(actor_network)}, but must be of type nn.Module."
        )


# Returns the expected action when given a state observation and epsilon=0 or 1.
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "image_space",
        "multidiscrete_space",
        "dict_space",
    ],
)
def test_returns_expected_action_epsilon_greedy(
    observation_space, discrete_space, request
):
    action_space = discrete_space
    observation_space = request.getfixturevalue(observation_space)

    dqn = DQN(observation_space, action_space)
    state = get_sample_from_space(observation_space)

    action_mask = None

    epsilon = 0
    action = dqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action >= 0 and action < action_space.n

    epsilon = 1
    action = dqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action >= 0 and action < action_space.n
    dqn.clean_up()


# Returns the expected action when given a state observation and action mask.
def test_returns_expected_action_mask(vector_space, discrete_space):
    accelerator = Accelerator()

    dqn = DQN(vector_space, discrete_space, accelerator=accelerator)
    state = get_sample_from_space(vector_space)

    action_mask = np.array([0, 1])

    epsilon = 0
    action = dqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action == 1

    epsilon = 1
    action = dqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action == 1
    dqn.clean_up()


def test_returns_expected_action_mask_vectorized(vector_space, discrete_space):
    accelerator = Accelerator()
    dqn = DQN(vector_space, discrete_space, accelerator=accelerator)
    state = get_sample_from_space(vector_space, batch_size=2)

    action_mask = np.array([[0, 1], [1, 0]])

    epsilon = 0
    action = dqn.get_action(state, epsilon, action_mask)

    assert np.array_equal(action, [1, 0])

    epsilon = 1
    action = dqn.get_action(state, epsilon, action_mask)

    assert np.array_equal(action, [1, 0])
    dqn.clean_up()


def test_dqn_optimizer_parameters(vector_space, discrete_space):
    dqn = DQN(vector_space, discrete_space)

    # Store initial parameters
    initial_params = {
        name: param.clone() for name, param in dqn.actor.named_parameters()
    }

    # Perform a dummy optimization step
    dummy_input = torch.randn(1, 4)
    dummy_return = torch.tensor([1.0])

    q_eval = dqn.actor(dummy_input)
    loss = (dummy_return - q_eval) ** 2
    loss = loss.mean()
    dqn.optimizer.zero_grad()
    loss.backward()
    dqn.optimizer.step()

    # Check if parameters have changed
    not_updated = []
    for name, param in dqn.actor.named_parameters():
        if torch.equal(initial_params[name], param):
            not_updated.append(name)

    assert not not_updated, f"The following parameters weren't updated:\n{not_updated}"
    dqn.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "image_space",
        "multidiscrete_space",
        "dict_space",
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("double", [False, True])
def test_learns_from_experiences(
    observation_space, accelerator, double, discrete_space, request
):
    action_space = discrete_space
    observation_space = request.getfixturevalue(observation_space)
    batch_size = 64

    # Create an instance of the DQN class
    dqn = DQN(
        observation_space,
        action_space,
        batch_size=batch_size,
        accelerator=accelerator,
        double=double,
    )

    # Create a batch of experiences
    device = accelerator.device if accelerator else "cpu"
    experiences = get_experiences_batch(
        observation_space, action_space, batch_size, device
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = dqn.actor
    actor_target = dqn.actor_target
    actor_pre_learn_sd = str(copy.deepcopy(dqn.actor.state_dict()))

    # Call the learn method
    loss = dqn.learn(experiences)

    assert isinstance(loss, float)
    assert loss >= 0.0
    assert actor == dqn.actor
    assert actor_target == dqn.actor_target
    assert actor_pre_learn_sd != str(dqn.actor.state_dict())
    dqn.clean_up()


# Updates target network parameters with soft update
def test_soft_update(vector_space, discrete_space):
    net_config = {"encoder_config": {"hidden_size": [64, 64]}}
    batch_size = 64
    lr = 1e-4
    learn_step = 5
    gamma = 0.99
    tau = 1e-3
    mut = None
    double = False
    actor_network = None
    device = "cpu"
    accelerator = None
    wrap = True

    dqn = DQN(
        vector_space,
        discrete_space,
        net_config=net_config,
        batch_size=batch_size,
        lr=lr,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        double=double,
        actor_network=actor_network,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
    )

    dqn.soft_update()

    eval_params = list(dqn.actor.parameters())
    target_params = list(dqn.actor_target.parameters())
    assert len(eval_params) == len(target_params)
    expected_params = [
        dqn.tau * eval_param + (1.0 - dqn.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]
    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )
    dqn.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
@pytest.mark.parametrize("num_envs", [1, 3])
def test_algorithm_test_loop(observation_space, discrete_space, num_envs, request):
    observation_space = request.getfixturevalue(observation_space)
    vect = num_envs > 1
    env = DummyEnv(observation_space=observation_space, vect=vect, num_envs=num_envs)
    agent = DQN(observation_space=observation_space, action_space=discrete_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
def test_clone_returns_identical_agent(vector_space, discrete_space):
    dqn = DummyDQN(vector_space, discrete_space)
    dqn.tensor_attribute = torch.randn(1)
    clone_agent = dqn.clone()

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    # assert clone_agent.actor_network == dqn.actor_network
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    assert clone_agent.tensor_attribute == dqn.tensor_attribute
    assert clone_agent.tensor_test == dqn.tensor_test
    dqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    dqn = DQN(vector_space, discrete_space, accelerator=accelerator)
    clone_agent = dqn.clone()

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    dqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    dqn = DQN(vector_space, discrete_space, accelerator=accelerator, wrap=False)
    clone_agent = dqn.clone(wrap=False)

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    dqn.clean_up()
    clone_agent.clean_up()


def test_clone_new_index(vector_space, discrete_space):
    dqn = DummyDQN(vector_space, discrete_space)
    clone_agent = dqn.clone(index=100)

    assert clone_agent.index == 100
    dqn.clean_up()
    clone_agent.clean_up()


def test_clone_after_learning(vector_space, discrete_space):
    batch_size = 8
    dqn = DQN(vector_space, discrete_space)

    states = torch.randn(batch_size, vector_space.shape[0])
    actions = torch.randint(0, 2, (batch_size, 1))
    rewards = torch.rand(batch_size, 1)
    next_states = torch.randn(batch_size, vector_space.shape[0])
    dones = torch.zeros(batch_size, 1)

    experiences = Transition(
        obs=states,
        action=actions,
        reward=rewards,
        next_obs=next_states,
        done=dones,
    ).to_tensordict()

    dqn.learn(experiences)
    clone_agent = dqn.clone()

    assert clone_agent.observation_space == dqn.observation_space
    assert clone_agent.action_space == dqn.action_space
    assert clone_agent.batch_size == dqn.batch_size
    assert clone_agent.lr == dqn.lr
    assert clone_agent.learn_step == dqn.learn_step
    assert clone_agent.gamma == dqn.gamma
    assert clone_agent.tau == dqn.tau
    assert clone_agent.mut == dqn.mut
    assert clone_agent.device == dqn.device
    assert clone_agent.accelerator == dqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), dqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), dqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == dqn.fitness
    assert clone_agent.steps == dqn.steps
    assert clone_agent.scores == dqn.scores
    dqn.clean_up()
    clone_agent.clean_up()



import h5py
import torch
import torch.nn as nn
import yaml

from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.training.train_offline import train_offline
from agilerl.utils.algo_utils import (
    get_input_size_from_space,
    get_output_size_from_space,
)
from agilerl.utils.utils import (
    create_population,
    make_vect_envs,
    observation_space_channels_to_first,
    print_hyperparams,
)
from agilerl.wrappers.make_evolvable import MakeEvolvable

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


class BasicNetActor(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size):
        super().__init__()
        layers = []

        # Add input layer
        layers.append(nn.Linear(input_size, hidden_sizes[0]))
        layers.append(nn.ReLU())  # Activation function

        # Add hidden layers
        for i in range(len(hidden_sizes) - 1):
            layers.append(nn.Linear(hidden_sizes[i], hidden_sizes[i + 1]))
            layers.append(nn.ReLU())  # Activation function

        # Add output layer with a sigmoid activation
        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        layers.append(nn.Softmax())  # Sigmoid activation

        # Combine all layers into a sequential model
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


def main(INIT_HP, MUTATION_PARAMS, NET_CONFIG):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("============ AgileRL ============")
    print(f"DEVICE: {device}")

    env = make_vect_envs(INIT_HP["ENV_NAME"], num_envs=INIT_HP["NUM_ENVS"])

    observation_space = env.single_observation_space
    action_space = env.single_action_space
    if INIT_HP["CHANNELS_LAST"]:
        observation_space = observation_space_channels_to_first(observation_space)

    state_dim = get_input_size_from_space(observation_space)
    action_dim = get_output_size_from_space(action_space)

    actor = BasicNetActor(state_dim[0], [32, 32], action_dim)
    actor_network = MakeEvolvable(
        actor, input_tensor=torch.ones(state_dim), device=device
    )

    dataset = h5py.File(INIT_HP["DATASET"], "r")

    memory = ReplayBuffer(INIT_HP["MEMORY_SIZE"], device=device)
    tournament = TournamentSelection(
        INIT_HP["TOURN_SIZE"],
        INIT_HP["ELITISM"],
        INIT_HP["POP_SIZE"],
        INIT_HP["EVAL_LOOP"],
    )
    mutations = Mutations(
        no_mutation=MUTATION_PARAMS["NO_MUT"],
        architecture=MUTATION_PARAMS["ARCH_MUT"],
        new_layer_prob=MUTATION_PARAMS["NEW_LAYER"],
        parameters=MUTATION_PARAMS["PARAMS_MUT"],
        activation=MUTATION_PARAMS["ACT_MUT"],
        rl_hp=MUTATION_PARAMS["RL_HP_MUT"],
        mutation_sd=MUTATION_PARAMS["MUT_SD"],
        min_lr=MUTATION_PARAMS["MIN_LR"],
        max_lr=MUTATION_PARAMS["MAX_LR"],
        min_batch_size=MUTATION_PARAMS["MAX_BATCH_SIZE"],
        max_batch_size=MUTATION_PARAMS["MAX_BATCH_SIZE"],
        rand_seed=MUTATION_PARAMS["RAND_SEED"],
        device=device,
    )

    agent_pop = create_population(
        algo=INIT_HP["ALGO"],
        observation_space=observation_space,
        action_space=action_space,
        net_config=NET_CONFIG,
        actor_network=actor_network,
        INIT_HP=INIT_HP,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=INIT_HP["NUM_ENVS"],
        device=device,
    )

    trained_pop, pop_fitnesses = train_offline(
        env,
        INIT_HP["ENV_NAME"],
        dataset,
        INIT_HP["ALGO"],
        agent_pop,
        memory=memory,
        INIT_HP=INIT_HP,
        MUT_P=MUTATION_PARAMS,
        swap_channels=INIT_HP["CHANNELS_LAST"],
        max_steps=INIT_HP["MAX_STEPS"],
        evo_steps=INIT_HP["EVO_STEPS"],
        eval_steps=INIT_HP["EVAL_STEPS"],
        eval_loop=INIT_HP["EVAL_LOOP"],
        target=INIT_HP["TARGET_SCORE"],
        tournament=tournament,
        mutation=mutations,
        wb=INIT_HP["WANDB"],
    )

    print_hyperparams(trained_pop)
    # plot_population_score(trained_pop)

    if str(device) == "cuda":
        torch.cuda.empty_cache()

    env.close()


if __name__ == "__main__":
    with open("configs/training/cqn.yaml") as file:
        cqn_config = yaml.safe_load(file)
    INIT_HP = cqn_config["INIT_HP"]
    MUTATION_PARAMS = cqn_config["MUTATION_PARAMS"]
    NET_CONFIG = cqn_config["NET_CONFIG"]
    # DISTRIBUTED_TRAINING = cqn_config["DISTRIBUTED_TRAINING"]
    main(INIT_HP, MUTATION_PARAMS, NET_CONFIG)



import random
from collections import deque, namedtuple
from numbers import Number
from typing import Any, Deque, NamedTuple, Optional, Union

import numpy as np
import torch

from agilerl.typing import NumpyObsType
from agilerl.utils.algo_utils import obs_to_tensor

NpTransitionType = Union[Number, np.ndarray, dict[str, np.ndarray]]
TorchTransitionType = Union[torch.Tensor, dict[str, torch.Tensor]]


class MultiAgentReplayBuffer:
    """The Multi-Agent Experience Replay Buffer class. Used to store multiple agents'
    experiences and allow off-policy learning.

    :param memory_size: Maximum length of the replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param agent_ids: Names of all agents that will act in the environment
    :type agent_ids: list[str]
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: Optional[str]
    """

    def __init__(
        self,
        memory_size: int,
        field_names: list[str],
        agent_ids: list[str],
        device: Optional[str] = None,
    ):
        assert memory_size > 0, "Memory size must be greater than zero."
        assert len(field_names) > 0, "Field names must contain at least one field name."
        assert len(agent_ids) > 0, "Agent ids must contain at least one agent id."

        self.memory_size: int = memory_size
        self.memory: Deque = deque(maxlen=memory_size)
        self.field_names: list[str] = field_names
        self.experience: NamedTuple = namedtuple(
            "Experience", field_names=self.field_names
        )
        self.counter: int = 0
        self.device: Optional[str] = device
        self.agent_ids: list[str] = agent_ids

    def __len__(self) -> int:
        """
        Returns the current size of internal memory.

        :return: Length of the memory
        :rtype: int
        """
        return len(self.memory)

    @staticmethod
    def stack_transitions(transitions: list[NumpyObsType]) -> NumpyObsType:
        """Stacks transitions into a single array/dictionary/tuple of arrays.

        :param transitions: List of transitions
        :type transitions: list[NumpyObsType]

        :return: Stacked transitions
        :rtype: NumpyObsType
        """
        # Identify the type of the transition
        field_type = type(transitions[0])

        # Stack the transitions into a single array or tuple/dictionary of arrays
        ts = []
        for ft in transitions:
            if field_type is dict:
                ft = {k: np.array(v) for k, v in ft.items()}
            elif field_type is tuple:
                ft = tuple(np.array(v) for v in ft)

            ts.append(ft)

        if field_type is dict:
            _ts = {}
            for k in ts[0].keys():
                kts = np.array([t[k] for t in ts])
                _ts[k] = np.expand_dims(kts, axis=1) if kts.ndim == 1 else kts

            ts = _ts
        elif field_type is tuple:
            _ts = []
            for i in range(len(ts[0])):
                its = np.array([t[i] for t in ts])
                _ts.append(np.expand_dims(its, axis=1) if its.ndim == 1 else its)

            ts = tuple(_ts)
        else:
            ts = np.array(ts)
            if ts.ndim == 1:
                ts = np.expand_dims(ts, axis=1)

        return ts

    def _add(self, *args: dict[str, NumpyObsType]) -> None:
        """
        Adds experience to memory.

        :param args: Variable length argument list for experience fields
        :type args: Any
        """
        e = self.experience(*args)
        self.memory.append(e)

    def _process_transition(
        self, experiences: list[NamedTuple], np_array: bool = False
    ) -> dict[str, dict[str, Any]]:
        """
        Returns transition dictionary from experiences.

        :param experiences: List of experiences
        :type experiences: list[NamedTuple]
        :param np_array: Flag to return numpy arrays instead of tensors, defaults to False
        :type np_array: bool, optional
        :return: Transition dictionary
        :rtype: dict[str, dict[str, Any]]
        """
        transition = {field: {} for field in self.field_names}
        experiences_filtered = [e for e in experiences if e is not None]

        for field in self.field_names:
            is_binary_field = field in [
                "done",
                "termination",
                "terminated",
                "truncation",
                "truncated",
            ]

            for agent_id in self.agent_ids:
                # Get field values for each agent
                ts = [getattr(e, field)[agent_id] for e in experiences_filtered]

                # Stack transitions if necessary
                ts = MultiAgentReplayBuffer.stack_transitions(ts)

                if is_binary_field and not np.isnan(ts).any():
                    ts = ts.astype(np.uint8)

                if not np_array:
                    ts = obs_to_tensor(ts, self.device)

                transition[field][agent_id] = ts

        return transition

    def sample(self, batch_size: int, *args: Any) -> tuple:
        """
        Returns sample of experiences from memory.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param args: Additional arguments
        :type args: Any
        :return: Sampled experiences
        :rtype: Tuple
        """
        experiences = random.sample(self.memory, k=batch_size)
        transition = self._process_transition(experiences)
        return tuple(transition.values())

    def save_to_memory_single_env(self, *args: dict[str, NumpyObsType]) -> None:
        """
        Saves experience to memory.

        :param args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        :type args: Any
        """
        self._add(*args)
        self.counter += 1

    def _reorganize_dicts(
        self, *args: dict[str, NumpyObsType]
    ) -> tuple[list[dict[str, NumpyObsType]], ...]:
        """
        Reorganizes dictionaries from vectorized to unvectorized experiences.

        :param args: Variable length argument list of dictionaries
        :type args: dict[str, np.ndarray]
        :return: Reorganized dictionaries
        :rtype: tuple[list[dict[str, np.ndarray]], ...]
        """

        def maybe_to_array(value):
            return np.array(value) if not isinstance(value, np.ndarray) else value

        results = [[] for _ in range(len(args))]
        num_entries = len(next(iter(args[0].values())))
        for i in range(num_entries):
            for j, arg in enumerate(args):
                new_dict = {}
                for key, value in arg.items():
                    if isinstance(value, dict):
                        new_dict[key] = {
                            k: maybe_to_array(v[i]) for k, v in value.items()
                        }
                    elif isinstance(value, tuple):
                        new_dict[key] = tuple(maybe_to_array(v[i]) for v in value)
                    else:
                        new_dict[key] = maybe_to_array(value[i])

                results[j].append(new_dict)

        return tuple(results)

    def save_to_memory_vect_envs(self, *args: dict[str, NumpyObsType]) -> None:
        """
        Saves multiple experiences to memory.

        :param args: Variable length argument list. Contains batched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        :type args: Any
        """
        args = self._reorganize_dicts(*args)
        for transition in zip(*args):
            self._add(*transition)
            self.counter += 1

    def save_to_memory(
        self, *args: dict[str, NumpyObsType], is_vectorised: bool = False
    ) -> None:
        """
        Applies appropriate save_to_memory function depending on whether
        the environment is vectorized or not.

        :param args: Variable length argument list. Contains batched or unbatched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        :type args: Any
        :param is_vectorised: Boolean flag indicating if the environment has been vectorized
        :type is_vectorised: bool
        """
        if is_vectorised:
            self.save_to_memory_vect_envs(*args)
        else:
            self.save_to_memory_single_env(*args)



# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Project information

project = "AgileRL"
copyright = "2023, AgileRL"
author = "Nick Ustaran-Anderegg"

release = "0.1"
version = "0.1.18"

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx_toolbox.collapse",
    "sphinx_github_changelog",
    "notfound.extension",
    "sphinxcontrib.googleanalytics",
]

googleanalytics_id = "G-ZKYJ8HJN69"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_title = "AgileRL Documentation"
html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_js_files = ["js/expand_sidebar.js"]
html_favicon = "_static/favicon.ico"
html_theme_options = {
    "sidebar_hide_name": False,
    "light_logo": "logo_teal.png",
    "dark_logo": "logo_white.png",
    "source_repository": "https://github.com/AgileRL/AgileRL/",
    "source_branch": "main",
    "source_directory": "docs/",
    "light_css_variables": {
        "color-brand-primary": "#245c5c",
        "color-brand-content": "#245c5c",
    },
    "dark_css_variables": {
        "color-brand-primary": "#48b8b8",
        "color-brand-content": "#48b8b8",
    },
}

# Configure sphinx-notfound-page
notfound_context = {
    "title": "Page Not Found",
    "body": """
<h1>Page Not Found</h1>
<p>Sorry, we couldn't find that page. Error code 404. </p>
<p>You can try using the search box or check our menu on the left hand side of this page.</p>

""",
}
notfound_no_urls_prefix = True

# -- Options for EPUB output
epub_show_urls = "footnote"



# Mock NetworkGroup for testing (avoids frame inspection issues)
from dataclasses import dataclass, field
from typing import Optional, Union

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim

from agilerl.algorithms.core.registry import (
    HyperparameterConfig,
    MutationRegistry,
    NetworkConfig,
    OptimizerConfig,
    RLParameter,
)
from agilerl.modules import EvolvableModule


@dataclass
class MockNetworkGroup:
    """Mock NetworkGroup for testing purposes."""

    eval_network: str
    shared_networks: Optional[Union[str, list[str]]] = field(default=None)
    policy: bool = field(default=False)

    def __hash__(self) -> int:
        return hash((self.eval_network, self.shared_networks, self.policy))


# Mock classes for testing
class MockEvolvableNetwork(EvolvableModule):
    """Mock evolvable network for testing purposes."""

    def __init__(self, input_dim=10, output_dim=5, name="mock_network", device="cpu"):
        super().__init__(device=device)
        self.name = name
        self.layer = nn.Linear(input_dim, output_dim)
        # Initialize with small weights for stable testing
        torch.nn.init.xavier_uniform_(self.layer.weight, gain=0.01)
        torch.nn.init.zeros_(self.layer.bias)

    def forward(self, x):
        return self.layer(x)

    def cpu(self):
        self.device = "cpu"
        return self

    def to(self, device):
        self.device = device
        return self

    def get_init_dict(self):
        return {"device": self.device}


class TestRLParameter:
    """Test suite for RLParameter class."""

    def test_rlparameter_init_defaults(self):
        """Test RLParameter initialization with default values."""
        param = RLParameter(min=0.1, max=2.0)

        assert param.min == 0.1
        assert param.max == 2.0
        assert param.shrink_factor == 0.8
        assert param.grow_factor == 1.2
        assert param.dtype == float
        assert param.value is None

    def test_rlparameter_init_custom(self):
        """Test RLParameter initialization with custom values."""
        param = RLParameter(
            min=1, max=10, shrink_factor=0.9, grow_factor=1.1, dtype=int
        )

        assert param.min == 1
        assert param.max == 10
        assert param.shrink_factor == 0.9
        assert param.grow_factor == 1.1
        assert param.dtype == int
        assert param.value is None

    def test_rlparameter_mutate_float(self):
        """Test mutation of float hyperparameters."""
        torch.manual_seed(42)  # For reproducible results

        param = RLParameter(min=0.1, max=2.0, dtype=float)
        param.value = 1.0

        # Test multiple mutations
        for _ in range(10):
            mutated = param.mutate()
            assert isinstance(mutated, float)
            assert param.min <= mutated <= param.max
            assert param.value == mutated

    def test_rlparameter_mutate_int(self):
        """Test mutation of integer hyperparameters."""
        torch.manual_seed(42)

        param = RLParameter(min=1, max=10, dtype=int)
        param.value = 5

        # Test multiple mutations
        for _ in range(10):
            mutated = param.mutate()
            assert isinstance(mutated, int)
            assert param.min <= mutated <= param.max
            assert param.value == mutated

    def test_rlparameter_mutate_numpy_array_float(self):
        """Test mutation of numpy float arrays."""
        torch.manual_seed(42)

        param = RLParameter(min=0.1, max=2.0, dtype=np.ndarray)
        param.value = np.array([0.5, 1.0, 1.5])

        original_dtype = param.value.dtype
        original_shape = param.value.shape

        # Test multiple mutations
        for _ in range(5):
            mutated = param.mutate()
            assert isinstance(mutated, np.ndarray)
            assert mutated.shape == original_shape
            assert mutated.dtype == original_dtype
            assert np.all(mutated >= param.min)
            assert np.all(mutated <= param.max)
            assert np.array_equal(param.value, mutated)

    def test_rlparameter_mutate_numpy_array_int(self):
        """Test mutation of numpy integer arrays."""
        torch.manual_seed(42)

        param = RLParameter(min=1, max=10, dtype=np.ndarray)
        param.value = np.array([3, 5, 7], dtype=np.int32)

        original_dtype = param.value.dtype
        original_shape = param.value.shape

        # Test multiple mutations
        for _ in range(5):
            mutated = param.mutate()
            assert isinstance(mutated, np.ndarray)
            assert mutated.shape == original_shape
            assert mutated.dtype == original_dtype
            assert np.all(mutated >= param.min)
            assert np.all(mutated <= param.max)
            assert np.array_equal(param.value, mutated)

    def test_rlparameter_mutate_multidimensional_array(self):
        """Test mutation of multi-dimensional numpy arrays."""
        torch.manual_seed(42)

        param = RLParameter(min=0.0, max=1.0, dtype=np.ndarray)
        param.value = np.array([[0.2, 0.4], [0.6, 0.8]])

        original_shape = param.value.shape

        # Test multiple mutations
        for _ in range(3):
            mutated = param.mutate()
            assert isinstance(mutated, np.ndarray)
            assert mutated.shape == original_shape
            assert np.all(mutated >= param.min)
            assert np.all(mutated <= param.max)

    def test_rlparameter_boundary_conditions(self):
        """Test mutation at boundary conditions."""
        param = RLParameter(
            min=0.0, max=1.0, shrink_factor=0.5, grow_factor=2.0, dtype=np.ndarray
        )

        # Test at minimum
        param.value = np.array([0.0, 0.0])
        mutated_min = param.mutate()
        assert np.all(mutated_min >= param.min)
        assert np.all(mutated_min <= param.max)

        # Test at maximum
        param.value = np.array([1.0, 1.0])
        mutated_max = param.mutate()
        assert np.all(mutated_max >= param.min)
        assert np.all(mutated_max <= param.max)

    def test_rlparameter_mutate_without_value_raises_error(self):
        """Test that mutation without setting value raises assertion error."""
        param = RLParameter(min=0.1, max=2.0)

        with pytest.raises(AssertionError, match="Hyperparameter value is not set"):
            param.mutate()


class TestHyperparameterConfig:
    """Test suite for HyperparameterConfig class."""

    def test_hyperparameter_config_init_empty(self):
        """Test initialization of empty HyperparameterConfig."""
        config = HyperparameterConfig()

        assert len(config.config) == 0
        assert not config  # Should be falsy when empty
        assert list(config.names()) == []

    def test_hyperparameter_config_init_with_params(self):
        """Test initialization with hyperparameters."""
        lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        batch_param = RLParameter(min=16, max=256, dtype=int)

        config = HyperparameterConfig(learning_rate=lr_param, batch_size=batch_param)

        assert len(config.config) == 2
        assert bool(config)  # Should be truthy when not empty
        assert set(config.names()) == {"learning_rate", "batch_size"}
        assert config["learning_rate"] == lr_param
        assert config["batch_size"] == batch_param

    def test_hyperparameter_config_invalid_param_raises_error(self):
        """Test that invalid parameter types raise ValueError."""
        with pytest.raises(ValueError, match="Expected RLParameter object"):
            HyperparameterConfig(invalid_param="not_an_rlparameter")

    def test_hyperparameter_config_iteration(self):
        """Test iteration over HyperparameterConfig."""
        lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        batch_param = RLParameter(min=16, max=256, dtype=int)

        config = HyperparameterConfig(learning_rate=lr_param, batch_size=batch_param)

        # Test iteration
        param_names = list(config)
        assert set(param_names) == {"learning_rate", "batch_size"}

        # Test items
        items = dict(config.items())
        assert items["learning_rate"] == lr_param
        assert items["batch_size"] == batch_param

    def test_hyperparameter_config_equality(self):
        """Test equality comparison between HyperparameterConfig objects."""
        lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        batch_param = RLParameter(min=16, max=256, dtype=int)

        config1 = HyperparameterConfig(learning_rate=lr_param, batch_size=batch_param)
        config2 = HyperparameterConfig(batch_size=batch_param, learning_rate=lr_param)
        config3 = HyperparameterConfig(learning_rate=lr_param)

        assert config1 == config2  # Same parameters, different order
        assert config1 != config3  # Different parameters

    def test_hyperparameter_config_sample(self):
        """Test sampling from HyperparameterConfig."""
        torch.manual_seed(42)

        lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        batch_param = RLParameter(min=16, max=256, dtype=int)

        config = HyperparameterConfig(learning_rate=lr_param, batch_size=batch_param)

        # Test sampling
        name, param = config.sample()
        assert name in ["learning_rate", "batch_size"]
        assert param in [lr_param, batch_param]

    def test_hyperparameter_config_repr(self):
        """Test string representation of HyperparameterConfig."""
        lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        config = HyperparameterConfig(learning_rate=lr_param)

        repr_str = repr(config)
        assert "HyperparameterConfig" in repr_str
        assert "learning_rate" in repr_str


class TestNetworkConfig:
    """Test suite for NetworkConfig class."""

    def test_network_config_init_eval_network(self):
        """Test NetworkConfig initialization for evaluation network."""
        config = NetworkConfig(name="actor", eval_network=True, optimizer="actor_opt")

        assert config.name == "actor"
        assert config.eval_network is True
        assert config.optimizer == "actor_opt"

    def test_network_config_init_shared_network(self):
        """Test NetworkConfig initialization for shared network."""
        config = NetworkConfig(name="target_actor", eval_network=False)

        assert config.name == "target_actor"
        assert config.eval_network is False
        assert config.optimizer is None

    def test_network_config_eval_without_optimizer_raises_error(self):
        """Test that evaluation network without optimizer raises ValueError."""
        with pytest.raises(
            ValueError, match="Evaluation network must have an optimizer"
        ):
            NetworkConfig(name="actor", eval_network=True, optimizer=None)


class TestOptimizerConfig:
    """Test suite for OptimizerConfig class."""

    def test_optimizer_config_init_single_network(self):
        """Test OptimizerConfig initialization with single network."""
        config = OptimizerConfig(
            name="actor_opt",
            networks="actor",
            lr="lr_actor",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={"weight_decay": 1e-4},
        )

        assert config.name == "actor_opt"
        assert config.networks == "actor"
        assert config.lr == "lr_actor"
        assert config.optimizer_cls == "Adam"  # Stored as string
        assert config.optimizer_kwargs == {"weight_decay": 1e-4}

    def test_optimizer_config_init_multiple_networks(self):
        """Test OptimizerConfig initialization with multiple networks."""
        config = OptimizerConfig(
            name="shared_opt",
            networks=["actor", "critic"],
            lr="lr_shared",
            optimizer_cls=optim.AdamW,
            optimizer_kwargs={"eps": 1e-8},
        )

        assert config.name == "shared_opt"
        assert config.networks == ["actor", "critic"]
        assert config.lr == "lr_shared"
        assert config.optimizer_cls == "AdamW"

    def test_optimizer_config_equality(self):
        """Test equality comparison between OptimizerConfig objects."""
        config1 = OptimizerConfig(
            name="opt1",
            networks="net1",
            lr="lr",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )
        config2 = OptimizerConfig(
            name="opt1",
            networks="net1",
            lr="lr_different",
            optimizer_cls=optim.SGD,
            optimizer_kwargs={"momentum": 0.9},
        )
        config3 = OptimizerConfig(
            name="opt2",
            networks="net1",
            lr="lr",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )

        assert config1 == config2  # Same name and networks
        assert config1 != config3  # Different name

    def test_optimizer_config_get_optimizer_cls(self):
        """Test getting optimizer class from stored configuration."""
        config = OptimizerConfig(
            name="test_opt",
            networks="test_net",
            lr="lr",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )

        optimizer_cls = config.get_optimizer_cls()
        assert optimizer_cls == optim.Adam

    def test_optimizer_config_get_optimizer_cls_dict(self):
        """Test getting optimizer classes from dict configuration."""
        # Create config with dict of optimizer classes
        config = OptimizerConfig(
            name="multi_opt",
            networks=["net1", "net2"],
            lr="lr",
            optimizer_cls={"agent1": optim.Adam, "agent2": optim.SGD},
            optimizer_kwargs={},
        )

        optimizer_dict = config.get_optimizer_cls()
        assert optimizer_dict["agent1"] == optim.Adam
        assert optimizer_dict["agent2"] == optim.SGD


class TestNetworkGroup:
    """Test suite for NetworkGroup class."""

    def setup_method(self):
        """Set up test fixtures."""
        self.mock_eval_net = MockEvolvableNetwork(name="eval_net")
        self.mock_shared_net = MockEvolvableNetwork(name="shared_net")
        self.mock_shared_nets = [
            MockEvolvableNetwork(name="shared_net1"),
            MockEvolvableNetwork(name="shared_net2"),
        ]

    def test_network_group_init_eval_only(self):
        """Test NetworkGroup initialization with evaluation network only."""
        # NetworkGroup uses frame inspection which doesn't work in test context
        # We'll test the make_network_group function instead which bypasses this
        pass

    def test_make_network_group_function(self):
        """Test the make_network_group helper function."""
        # Create a simple NetworkGroup directly with string names to avoid frame inspection
        from dataclasses import dataclass, field
        from typing import Optional, Union

        # Create a simplified NetworkGroup for testing
        @dataclass
        class SimpleNetworkGroup:
            eval_network: str
            shared_networks: Optional[Union[str, list[str]]] = field(default=None)
            policy: bool = field(default=False)

        # Test the basic structure
        group = SimpleNetworkGroup(
            eval_network="actor", shared_networks="target_actor", policy=True
        )

        assert group.eval_network == "actor"
        assert group.shared_networks == "target_actor"
        assert group.policy is True

    def test_make_network_group_with_list(self):
        """Test make_network_group with list of shared networks."""
        from dataclasses import dataclass, field
        from typing import Optional, Union

        @dataclass
        class SimpleNetworkGroup:
            eval_network: str
            shared_networks: Optional[Union[str, list[str]]] = field(default=None)
            policy: bool = field(default=False)

        group = SimpleNetworkGroup(
            eval_network="critic",
            shared_networks=["target_critic1", "target_critic2"],
            policy=False,
        )

        assert group.eval_network == "critic"
        assert group.shared_networks == ["target_critic1", "target_critic2"]
        assert group.policy is False

    def test_make_network_group_no_shared(self):
        """Test make_network_group with no shared networks."""
        from dataclasses import dataclass, field
        from typing import Optional, Union

        @dataclass
        class SimpleNetworkGroup:
            eval_network: str
            shared_networks: Optional[Union[str, list[str]]] = field(default=None)
            policy: bool = field(default=False)

        group = SimpleNetworkGroup(
            eval_network="policy", shared_networks=None, policy=True
        )

        assert group.eval_network == "policy"
        assert group.shared_networks is None
        assert group.policy is True


class TestMutationRegistry:
    """Test suite for MutationRegistry class."""

    def setup_method(self):
        """Set up test fixtures."""
        self.lr_param = RLParameter(min=1e-5, max=1e-2, dtype=float)
        self.batch_param = RLParameter(min=16, max=256, dtype=int)
        self.hp_config = HyperparameterConfig(
            learning_rate=self.lr_param, batch_size=self.batch_param
        )

    def test_mutation_registry_init_empty(self):
        """Test MutationRegistry initialization without hyperparameters."""
        registry = MutationRegistry()

        assert registry.hp_config is not None
        assert len(registry.hp_config.config) == 0
        assert len(registry.groups) == 0
        assert len(registry.optimizers) == 0
        assert len(registry.hooks) == 0

    def test_mutation_registry_init_with_hp_config(self):
        """Test MutationRegistry initialization with hyperparameters."""
        registry = MutationRegistry(hp_config=self.hp_config)

        assert registry.hp_config == self.hp_config
        assert len(registry.groups) == 0
        assert len(registry.optimizers) == 0

    def test_mutation_registry_register_group(self):
        """Test registering network groups."""
        registry = MutationRegistry()

        group1 = MockNetworkGroup("actor", "target_actor", policy=True)
        group2 = MockNetworkGroup("critic", None, policy=False)

        registry.register_group(group1)
        registry.register_group(group2)

        assert len(registry.groups) == 2
        assert group1 in registry.groups
        assert group2 in registry.groups

    def test_mutation_registry_register_optimizer(self):
        """Test registering optimizers."""
        registry = MutationRegistry()

        opt_config = OptimizerConfig(
            name="actor_opt",
            networks="actor",
            lr="lr_actor",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )

        registry.register_optimizer(opt_config)

        assert len(registry.optimizers) == 1
        assert registry.optimizers[0] == opt_config

    def test_mutation_registry_register_hook(self):
        """Test registering mutation hooks."""
        registry = MutationRegistry()

        def dummy_hook():
            pass

        registry.register_hook(dummy_hook)

        assert len(registry.hooks) == 1
        assert registry.hooks[0] == "dummy_hook"

    def test_mutation_registry_policy_method(self):
        """Test getting policy network from registry."""
        registry = MutationRegistry()

        # No policy registered
        assert registry.policy() is None

        # Register non-policy group
        group1 = MockNetworkGroup("critic", None, policy=False)
        registry.register_group(group1)
        assert registry.policy() is None

        # Register policy group
        group2 = MockNetworkGroup("actor", "target_actor", policy=True)
        registry.register_group(group2)
        assert registry.policy() == "actor"
        assert registry.policy(return_group=True) == group2

    def test_mutation_registry_all_registered(self):
        """Test getting all registered components."""
        registry = MutationRegistry()

        # Register groups
        group1 = MockNetworkGroup("actor", "target_actor", policy=True)
        group2 = MockNetworkGroup(
            "critic", ["target_critic1", "target_critic2"], policy=False
        )
        registry.register_group(group1)
        registry.register_group(group2)

        # Register optimizer
        opt_config = OptimizerConfig(
            name="actor_opt",
            networks="actor",
            lr="lr_actor",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )
        registry.register_optimizer(opt_config)

        all_registered = registry.all_registered()
        expected = {
            "actor",
            "target_actor",
            "critic",
            "target_critic1",
            "target_critic2",
            "actor_opt",
        }
        assert set(all_registered) == expected

    def test_mutation_registry_networks(self):
        """Test getting network configurations."""
        registry = MutationRegistry()

        # Register groups
        group1 = MockNetworkGroup("actor", "target_actor", policy=True)
        group2 = MockNetworkGroup("critic", None, policy=False)
        registry.register_group(group1)
        registry.register_group(group2)

        # Register optimizers
        actor_opt_config = OptimizerConfig(
            name="actor_opt",
            networks=["actor"],
            lr="lr_actor",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )
        critic_opt_config = OptimizerConfig(
            name="critic_opt",
            networks=["critic"],
            lr="lr_critic",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )
        registry.register_optimizer(actor_opt_config)
        registry.register_optimizer(critic_opt_config)

        networks = registry.networks()

        # Should have 3 networks: actor (eval), target_actor (shared), critic (eval)
        assert len(networks) == 3

        # Check eval networks
        eval_networks = [net for net in networks if net.eval_network]
        assert len(eval_networks) == 2

        # Check shared networks
        shared_networks = [net for net in networks if not net.eval_network]
        assert len(shared_networks) == 1

    def test_mutation_registry_optimizer_networks_property(self):
        """Test optimizer_networks property."""
        registry = MutationRegistry()

        opt1 = OptimizerConfig(
            name="opt1",
            networks="net1",
            lr="lr",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )
        opt2 = OptimizerConfig(
            name="opt2",
            networks=["net2", "net3"],
            lr="lr",
            optimizer_cls=optim.SGD,
            optimizer_kwargs={},
        )

        registry.register_optimizer(opt1)
        registry.register_optimizer(opt2)

        optimizer_networks = registry.optimizer_networks
        assert optimizer_networks["opt1"] == "net1"
        assert optimizer_networks["opt2"] == ["net2", "net3"]

    def test_mutation_registry_equality(self):
        """Test equality comparison between MutationRegistry objects."""
        registry1 = MutationRegistry(hp_config=self.hp_config)
        registry2 = MutationRegistry(hp_config=self.hp_config)
        registry3 = MutationRegistry()

        # Add same groups and optimizers to registry1 and registry2
        group = MockNetworkGroup("actor", None, policy=True)
        opt = OptimizerConfig(
            name="opt",
            networks="actor",
            lr="lr",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )

        registry1.register_group(group)
        registry1.register_optimizer(opt)
        registry2.register_group(group)
        registry2.register_optimizer(opt)

        assert registry1 == registry2
        assert registry1 != registry3

    def test_mutation_registry_repr(self):
        """Test string representation of MutationRegistry."""
        registry = MutationRegistry()

        group = MockNetworkGroup("actor", "target_actor", policy=True)
        opt = OptimizerConfig(
            name="actor_opt",
            networks="actor",
            lr="lr_actor",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={},
        )

        registry.register_group(group)
        registry.register_optimizer(opt)

        repr_str = repr(registry)
        assert "Network Groups:" in repr_str
        assert "Optimizers:" in repr_str
        assert "actor" in repr_str
        assert "target_actor" in repr_str
        assert "actor_opt" in repr_str


class TestIntegration:
    """Integration tests for registry components working together."""

    def test_complete_registry_setup(self):
        """Test complete registry setup with all components."""
        # Create hyperparameter configuration
        hp_config = HyperparameterConfig(
            learning_rate=RLParameter(min=1e-5, max=1e-2, dtype=float),
            batch_size=RLParameter(min=16, max=256, dtype=int),
            layer_sizes=RLParameter(min=32, max=512, dtype=np.ndarray),
        )

        # Create registry
        registry = MutationRegistry(hp_config=hp_config)

        # Register network groups
        actor_group = MockNetworkGroup("actor", "target_actor", policy=True)
        critic_group = MockNetworkGroup("critic", "target_critic", policy=False)
        registry.register_group(actor_group)
        registry.register_group(critic_group)

        # Register optimizers
        actor_opt = OptimizerConfig(
            name="actor_opt",
            networks=["actor"],
            lr="learning_rate",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={"weight_decay": 1e-4},
        )
        critic_opt = OptimizerConfig(
            name="critic_opt",
            networks=["critic"],
            lr="learning_rate",
            optimizer_cls=optim.Adam,
            optimizer_kwargs={"weight_decay": 1e-4},
        )
        registry.register_optimizer(actor_opt)
        registry.register_optimizer(critic_opt)

        # Register hook
        def post_mutation_hook():
            pass

        registry.register_hook(post_mutation_hook)

        # Verify complete setup
        assert len(registry.hp_config.config) == 3
        assert len(registry.groups) == 2
        assert len(registry.optimizers) == 2
        assert len(registry.hooks) == 1

        # Verify policy detection
        assert registry.policy() == "actor"

        # Verify all components are registered
        all_registered = registry.all_registered()
        expected = {
            "actor",
            "target_actor",
            "critic",
            "target_critic",
            "actor_opt",
            "critic_opt",
        }
        assert set(all_registered) == expected

        # Verify network configurations
        networks = registry.networks()
        eval_networks = [net for net in networks if net.eval_network]
        shared_networks = [net for net in networks if not net.eval_network]

        assert len(eval_networks) == 2  # actor and critic
        assert len(shared_networks) == 2  # target_actor and target_critic

        # Verify optimizer associations
        actor_net_config = next(net for net in eval_networks if net.name == "actor")
        critic_net_config = next(net for net in eval_networks if net.name == "critic")

        assert actor_net_config.optimizer == "actor_opt"
        assert critic_net_config.optimizer == "critic_opt"

    def test_numpy_array_hyperparameter_mutation_integration(self):
        """Test integration of numpy array hyperparameters with registry."""
        torch.manual_seed(42)

        # Create numpy array hyperparameters
        layer_sizes_param = RLParameter(min=16, max=512, dtype=np.ndarray)
        layer_sizes_param.value = np.array([128, 64, 32], dtype=np.int32)

        dropout_rates_param = RLParameter(min=0.0, max=0.5, dtype=np.ndarray)
        dropout_rates_param.value = np.array([0.1, 0.2, 0.3])

        hp_config = HyperparameterConfig(
            layer_sizes=layer_sizes_param, dropout_rates=dropout_rates_param
        )

        # Test that we can sample and mutate numpy array hyperparameters
        for _ in range(5):
            param_name, param = hp_config.sample()
            assert param_name in ["layer_sizes", "dropout_rates"]

            original_value = param.value.copy()
            mutated_value = param.mutate()

            assert isinstance(mutated_value, np.ndarray)
            assert mutated_value.shape == original_value.shape
            assert np.all(mutated_value >= param.min)
            assert np.all(mutated_value <= param.max)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
import torch
from accelerate import Accelerator
from tensordict import TensorDict
from torch.utils.data import DataLoader

from agilerl.components.data import ReplayDataset
from agilerl.components.replay_buffer import (
    MultiStepReplayBuffer,
    PrioritizedReplayBuffer,
    ReplayBuffer,
)
from agilerl.components.sampler import Sampler


# Initialize Sampler with default parameters
def test_initialize_with_default_parameters():
    memory_size = 100
    device = "cpu"
    buffer = ReplayBuffer(memory_size, device)

    sampler = Sampler(memory=buffer)
    assert sampler.distributed is False
    assert sampler.per is False
    assert sampler.n_step is False
    assert sampler.memory == buffer
    assert sampler.dataset is None
    assert sampler.dataloader is None


# Call sample_standard() method with valid batch_size
def test_sample_standard_with_valid_batch_size():
    memory_size = 100
    device = "cpu"

    buffer = ReplayBuffer(memory_size, device)
    sampler = Sampler(
        memory=buffer,
        dataset=None,
        dataloader=None,
    )

    experiences1 = TensorDict(
        {
            "state": [1],
            "action": [2],
            "reward": [3],
        },
        batch_size=[1],
    )
    experiences2 = TensorDict(
        {
            "state": [4],
            "action": [5],
            "reward": [6],
        },
        batch_size=[1],
    )
    experiences3 = TensorDict(
        {
            "state": [7],
            "action": [8],
            "reward": [9],
        },
        batch_size=[1],
    )
    # Add experiences to memory
    buffer.add(experiences1)
    buffer.add(experiences2)
    buffer.add(experiences3)

    batch_size = 3
    samples = sampler.sample(batch_size)

    # TensorDict should have the right shape
    assert samples.batch_size[0] == batch_size
    # Check that all keys are present
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples
    # Check that tensors have the expected batch dimension
    assert samples["state"].shape[0] == batch_size


# Call sample_distributed() method with valid batch_size
def test_sample_distributed_with_valid_batch_size():
    accelerator = Accelerator()

    memory_size = 100
    batch_size = 3

    buffer = ReplayBuffer(memory_size)
    replay_dataset = ReplayDataset(buffer, batch_size=batch_size)

    replay_dataloader = DataLoader(replay_dataset, batch_size=None)
    replay_dataloader = accelerator.prepare(replay_dataloader)
    sampler = Sampler(
        dataset=replay_dataset,
        dataloader=replay_dataloader,
    )

    # Add experiences to memory
    experiences1 = TensorDict(
        {
            "state": [1],
            "action": [2],
            "reward": [3],
        },
        batch_size=[1],
    )
    experiences2 = TensorDict(
        {
            "state": [4],
            "action": [5],
            "reward": [6],
        },
        batch_size=[1],
    )
    experiences3 = TensorDict(
        {
            "state": [7],
            "action": [8],
            "reward": [9],
        },
        batch_size=[1],
    )
    buffer.add(experiences1)
    buffer.add(experiences2)
    buffer.add(experiences3)

    samples = sampler.sample(batch_size)

    # Check that all expected keys are present
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples

    # Check that tensors have the expected dimensions by examining their shape directly
    assert samples["state"].shape[0] == batch_size


# Call sample_per() method with valid batch_size
def test_sample_per_with_valid_batch_size():
    memory_size = 100
    alpha = 0.6
    device = "cpu"

    buffer = PrioritizedReplayBuffer(memory_size, alpha, device)
    sampler = Sampler(
        memory=buffer,
        dataset=None,
        dataloader=None,
    )

    # Add experiences to memory
    experiences1 = TensorDict(
        {
            "state": torch.tensor([[1]], dtype=torch.float32),
            "action": torch.tensor([[2]], dtype=torch.float32),
            "reward": torch.tensor([[3]], dtype=torch.float32),
            "next_state": torch.tensor([[2]], dtype=torch.float32),
            "done": torch.tensor([[1]], dtype=torch.float32),
        },
        batch_size=[1],
    )
    experiences2 = TensorDict(
        {
            "state": torch.tensor([[4]], dtype=torch.float32),
            "action": torch.tensor([[5]], dtype=torch.float32),
            "reward": torch.tensor([[6]], dtype=torch.float32),
            "next_state": torch.tensor([[5]], dtype=torch.float32),
            "done": torch.tensor([[0]], dtype=torch.float32),
        },
        batch_size=[1],
    )
    experiences3 = TensorDict(
        {
            "state": torch.tensor([[7]], dtype=torch.float32),
            "action": torch.tensor([[8]], dtype=torch.float32),
            "reward": torch.tensor([[9]], dtype=torch.float32),
            "next_state": torch.tensor([[8]], dtype=torch.float32),
            "done": torch.tensor([[0]], dtype=torch.float32),
        },
        batch_size=[1],
    )

    buffer.add(experiences1)
    buffer.add(experiences2)
    buffer.add(experiences3)

    batch_size = 2  # Reduced batch size to avoid dimension mismatch issues
    samples = sampler.sample(batch_size, beta=0.4)

    # Check that expected keys are present
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples
    assert "next_state" in samples
    assert "done" in samples
    assert "weights" in samples
    assert "idxs" in samples

    # Check dimensions match the requested batch size
    assert samples["weights"].shape[0] == batch_size
    assert samples["idxs"].shape[0] == batch_size


# Call sample_n_step() method with valid batch_size
def test_sample_n_step_with_valid_batch_size():
    memory_size = 10000
    n_step = 3
    gamma = 0.95
    device = "cpu"

    buffer = MultiStepReplayBuffer(memory_size, n_step, gamma, device)
    sampler = Sampler(
        memory=buffer,
        dataset=None,
        dataloader=None,
    )

    # Add experiences to memory
    experiences1 = TensorDict(
        {
            "obs": [1],
            "action": [2],
            "reward": [3.0],
            "next_obs": [2],
            "done": [1],
        },
    )
    experiences2 = TensorDict(
        {
            "obs": [4],
            "action": [5],
            "reward": [6.0],
            "next_obs": [5],
            "done": [0],
        },
    )
    experiences3 = TensorDict(
        {
            "obs": [7],
            "action": [8],
            "reward": [9.0],
            "next_obs": [8],
            "done": [0],
        },
    )
    experiences4 = TensorDict(
        {
            "obs": [9],
            "action": [8],
            "reward": [7.0],
            "next_obs": [8],
            "done": [0],
        },
    )
    for experience in [experiences1, experiences2, experiences3, experiences4]:
        experience = experience.unsqueeze(0)
        experience.batch_size = [1]
        buffer.add(experience)

    idxs = torch.tensor([0, 1])
    samples = sampler.sample(idxs)

    # Check that some expected keys are present
    assert "obs" in samples
    assert "action" in samples
    assert "reward" in samples

    # Check dimensions
    assert samples["obs"].shape[0] == len(idxs)


@pytest.mark.parametrize(
    "memory, dataset, dataloader",
    [
        (0, None, None),
        (None, 0, 0),
        (0, None, None),
        (0, None, None),
    ],
)
def test_warnings_in_constructor(memory, dataset, dataloader):
    with pytest.warns():
        _ = Sampler(
            memory,
            dataset,
            dataloader,
        )


# Test that the Sampler correctly replaces the collate function in a regular DataLoader
def test_replace_dataloader_collate_fn():
    memory_size = 100
    batch_size = 3

    buffer = ReplayBuffer(memory_size)
    replay_dataset = ReplayDataset(buffer, batch_size=batch_size)

    # Create a regular DataLoader with default collate function
    original_dataloader = DataLoader(replay_dataset, batch_size=None)

    # Create a Sampler that should replace the collate function
    sampler = Sampler(
        memory=None,
        dataset=replay_dataset,
        dataloader=original_dataloader,
    )

    # Add experiences to memory
    experiences1 = TensorDict(
        {
            "state": [1],
            "action": [2],
            "reward": [3],
        },
        batch_size=[1],
    )
    experiences2 = TensorDict(
        {
            "state": [4],
            "action": [5],
            "reward": [6],
        },
        batch_size=[1],
    )
    experiences3 = TensorDict(
        {
            "state": [7],
            "action": [8],
            "reward": [9],
        },
        batch_size=[1],
    )
    buffer.add(experiences1)
    buffer.add(experiences2)
    buffer.add(experiences3)

    # Verify that the Sampler created a new dataloader and stored the original
    assert sampler.dataloader is not original_dataloader
    assert sampler.dataloader.collate_fn == Sampler.tensordict_collate_fn

    # Sample from the sampler
    samples = sampler.sample(batch_size)

    # Check that all expected keys are present
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples

    # Check that tensors have the expected dimensions
    assert samples["state"].shape[0] == batch_size



import gc
from typing import Any, Optional

import numpy as np
import torch
from accelerate import Accelerator

from agilerl import HAS_LLM_DEPENDENCIES
from agilerl.algorithms.core import LLMAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, NetworkGroup
from agilerl.protocols import (
    LoraConfigProtocol,
    PeftModelProtocol,
    PreTrainedModelProtocol,
)
from agilerl.typing import ExperiencesType, LLMObsType
from agilerl.utils.algo_utils import (
    CosineLRScheduleConfig,
    VLLMConfig,
    get_experiences_samples,
    stack_and_pad_experiences,
)
from agilerl.utils.llm_utils import (
    ReasoningGym,
)

if HAS_LLM_DEPENDENCIES:
    from transformers import GenerationConfig


class GRPO(LLMAlgorithm):
    """The GRPO algorithm class. GRPO paper: https://arxiv.org/pdf/2402.03300

    :param pad_token_id: Pad token id
    :type pad_token_id: int
    :param pad_token: Pad token
    :type pad_token: str
    :param model_name: Model name
    :type model_name: str, optional
    :param actor_network: HuggingFace LLM
    :type actor_network: PreTrainedModelProtocol
    :param model_config: Model configuration, to be used when creating the model from a name or path
    :type model_config: dict[str, Any], optional
    :param hp_config: RL hyperparameter mutation configuration, defaults to None, whereby algorithm mutations are disabled.
    :type hp_config: HyperparameterConfig, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param batch_size: Mini-batch size for learning, defaults to 64
    :type batch_size: int, optional
    :param beta: Beta coefficient, controls the strength of the KL divergence penalty, defaults to 0.001
    :type beta: float, optional
    :param lr: Learning rate for optimizer, defaults to 1e-4
    :type lr: float, optional
    :param clip_coef: Surrogate clipping coefficient, defaults to 0.2
    :type clip_coef: float, optional
    :param max_grad_norm: Maximum norm for gradient clipping, defaults to 0.1
    :type max_grad_norm: float, optional
    :param update_epochs: Number of policy update epochs, defaults to 4
    :type update_epochs: int, optional
    :param group_size: Group size, defaults to 8
    :type group_size: int, optional
    :param temperature: Temperature, controls randomness of text generation
    :type temperature: float, optional
    :param calc_position_embeddings: Flag indicating whether to calculate position embeddings, defaults to True
    :type calc_position_embeddings: bool, optional
    :param reduce_memory_peak: Flag to reduce memory peak in the _get_logprobs method, defaults to False
    :type reduce_memory_peak: bool, optional
    :param micro_batch_size_per_gpu: If specified, gradient_accumulation_steps will be
        calculated to achieve the target batch_size. If None, uses existing
        gradient_accumulation_steps from DeepSpeed config, defaults to None
    :type micro_batch_size_per_gpu: int, optional
    :param max_output_tokens: Max number of answer tokens, defaults to 512
    :type max_output_tokens: int, optional
    :param min_output_tokens: Minimum output tokens, defaults to 0
    :type min_output_tokens: int, optional
    :param max_model_len: Maximum context window length, defaults to None
    :type max_model_len: int, optional
    :param lora_config: Config for LoRA, defaults to None
    :type lora_config: LoraConfigProtocol, optional
    :param cosine_lr_schedule_config: Config for cosine lr scheduling, defaults to None
    :type cosine_lr_schedule_config: CosineLRScheduleConfig, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    :param clone: Flag to indicate if the instantiation is a cloning, defaults to False
    :type clone: bool, optional
    :param use_separate_reference_adapter: Flag to indicate if the reference policy should have a separate adapter, defaults to False
    :type use_separate_reference_adapter: bool, optional
    :param use_vllm: Flag to indicate if the model should use vllm for generation, defaults to False
    :type use_vllm: bool, optional
    :param vllm_config: Config for VLLM generation, defaults to None
    :type vllm_config: VLLMConfig, optional
    :param seed: Seed for the random number generator, defaults to 42
    :type seed: int, optional
    :param gradient_checkpointing: Flag to indicate if gradient checkpointing should be used, defaults to True
    :type gradient_checkpointing: bool, optional
    """

    def __init__(
        self,
        pad_token_id: int,
        pad_token: str,
        model_name: str | None = None,
        actor_network: PreTrainedModelProtocol | None = None,
        model_config: dict[str, Any] | None = None,
        hp_config: Optional[HyperparameterConfig] = None,
        index: int = 0,
        batch_size: int = 16,
        beta: float = 0.001,
        lr: float = 5e-7,
        clip_coef: float = 0.2,
        max_grad_norm: float = 0.1,
        update_epochs: int = 1,
        group_size: int = 8,
        temperature: float = 0.9,
        repetition_penalty: float = 1.0,
        top_p: float = 0.95,
        top_k: int = 50,
        min_p: float = 0.0,
        calc_position_embeddings: bool = True,
        micro_batch_size_per_gpu: int | None = None,
        reduce_memory_peak: bool = False,
        max_output_tokens: int | None = 1024,
        min_output_tokens: Optional[int] = None,
        max_model_len: Optional[int] = None,
        lora_config: Optional[LoraConfigProtocol] = None,
        cosine_lr_schedule_config: Optional[CosineLRScheduleConfig] = None,
        accelerator: Optional[Accelerator] = None,
        device: str = "cpu",
        wrap: bool = True,
        clone: bool = False,
        use_separate_reference_adapter: bool = False,
        use_vllm: bool = False,
        vllm_config: Optional[VLLMConfig] = None,
        seed: int = 42,
        gradient_checkpointing: bool = True,
    ) -> None:

        device = (
            f"cuda:{accelerator.process_index}"
            if accelerator is not None
            else ("cuda" if torch.cuda.is_available() else "cpu")
        )
        super().__init__(
            index=index,
            batch_size=batch_size,
            lr=lr,
            max_grad_norm=max_grad_norm,
            clone=clone,
            reduce_memory_peak=reduce_memory_peak,
            calc_position_embeddings=calc_position_embeddings,
            seed=seed,
            pad_token_id=pad_token_id,
            pad_token=pad_token,
            lora_config=lora_config,
            use_separate_reference_adapter=use_separate_reference_adapter,
            model_name=model_name,
            actor_network=actor_network,
            model_config=model_config,
            micro_batch_size_per_gpu=micro_batch_size_per_gpu,
            cosine_lr_schedule_config=cosine_lr_schedule_config,
            wrap=wrap,
            hp_config=hp_config,
            device=device,
            accelerator=accelerator,
            name="GRPO",
            gradient_checkpointing=gradient_checkpointing,
        )
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(
            clip_coef, (float, int)
        ), "Clipping coefficient must be a float."
        assert (
            clip_coef >= 0
        ), "Clipping coefficient must be greater than or equal to zero."
        assert isinstance(
            update_epochs, int
        ), "Policy update epochs must be an integer."
        assert (
            update_epochs >= 1
        ), "Policy update epochs must be greater than or equal to one."
        if actor_network is not None:
            assert isinstance(
                actor_network, (PeftModelProtocol, PreTrainedModelProtocol)
            ), "Actor network must be a PeftModelProtocol or PreTrainedModelProtocol"

        self.clip_coef = clip_coef
        self.update_epochs = update_epochs
        self.group_size = group_size
        self.beta = beta
        self.temperature = temperature
        self.repetition_penalty = repetition_penalty
        self.top_p = top_p
        self.top_k = top_k
        self.min_p = min_p
        if max_output_tokens is None and max_model_len is None:
            raise ValueError(
                "Either max_output_tokens or max_model_len must be specified"
            )
        self.max_output_tokens = max_output_tokens
        self.min_output_tokens = min_output_tokens
        self.max_model_len = (
            max_model_len if max_model_len is not None else max_output_tokens + 512
        )
        self.generation_config = GenerationConfig(
            do_sample=True,
            temperature=temperature,
            max_length=self.max_model_len,
            max_new_tokens=max_output_tokens,
            min_new_tokens=min_output_tokens,
            pad_token_id=pad_token_id,
            repetition_penalty=repetition_penalty,
            top_p=top_p,
            top_k=top_k,
            min_p=min_p,
        )

        self.use_vllm = use_vllm
        self.vllm_config = vllm_config
        if self.use_vllm:
            self._configure_vllm()
        self._initialize_actors(actor_network, not clone)
        # Register network groups for mutations
        self.register_network_group(NetworkGroup(eval_network=self.actor, policy=True))
        if self.wrap:
            self.wrap_models()

    def get_action(
        self, obs: LLMObsType, training: bool = True
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """Returns the next action to take in the environment.

        :param obs: Environment observation, or multiple observations in a batch
        :type obs: numpy.ndarray[float]
        :param training: Flag to indicate training mode, defaults to True
        :type training: bool, optional
        :return: Completion IDs and action masks
        :rtype: tuple[list[torch.Tensor], list[torch.Tensor]]
        """
        group_size = self.group_size if training else 1
        self.actor.eval()
        if not self.use_vllm:
            with torch.no_grad():
                completion_ids = []
                action_masks = []
                for prompt in obs:
                    prompt.pop("text", None)
                    prompt["input_ids"] = (
                        prompt["input_ids"].repeat(group_size, 1).to(self.actor.device)
                    )
                    prompt["attention_mask"] = (
                        prompt["attention_mask"]
                        .repeat(group_size, 1)
                        .to(self.actor.device)
                    )
                    completion_id = self.actor.generate(
                        **prompt,
                        generation_config=self.generation_config,
                    )
                    completion_ids.append(completion_id)
                    action_mask = torch.zeros_like(
                        completion_id, dtype=torch.bool, device=self.device
                    )
                    action_mask[:, prompt["input_ids"].shape[1] :] = True
                    action_mask[completion_id == self.pad_token_id] = False
                    action_mask = action_mask[:, 1:]
                    action_masks.append(action_mask)
        else:
            if self.vllm_config.sleep_mode:
                torch.cuda.empty_cache()
                self.llm.wake_up()
            self._move_model_to_vllm()
            completion_ids, action_masks = self._generate_with_vllm_colocate(
                obs, group_size
            )
            if self.vllm_config.sleep_mode:
                self.llm.sleep(level=2)

        return completion_ids, action_masks

    def learn(self, experiences: ExperiencesType) -> tuple[float, float]:
        """Updates agent network parameters to learn from experiences.

        :param experiences: Batched completion_ids, action_masks and rewards
        :type experiences: ExperiencesType
        """
        gc.collect()
        torch.cuda.empty_cache()

        completion_ids, action_masks, rewards = stack_and_pad_experiences(
            *experiences, padding_values=[self.pad_token_id, False, None]
        )
        advantages = self._calculate_advantage(rewards).to(self.device)

        num_samples = advantages.shape[0]
        batch_idxs = np.arange(num_samples)
        mean_loss, mean_kl = 0, 0
        batch_size = min(num_samples, self.micro_batch_size_per_gpu)

        with torch.no_grad():
            reference_log_probs = self._get_logprobs(
                completion_ids,
                batch_size=batch_size,
                use_reference=True,
                eval_mode=True,
            )
            old_log_probs = self._get_logprobs(
                completion_ids,
                batch_size=batch_size,
                use_reference=False,
                eval_mode=True,
            )

        experiences = (
            completion_ids,
            action_masks,
            advantages,
            old_log_probs,
            reference_log_probs,
        )

        for _ in range(self.update_epochs):
            self.rng.shuffle(batch_idxs)
            for start in range(0, num_samples, batch_size):
                minibatch_idxs = batch_idxs[
                    start : min((start + batch_size), num_samples)
                ]
                (
                    batch_ids,
                    batch_action_mask,
                    batch_advantages,
                    batch_old_log_probs,
                    batch_reference_log_probs,
                ) = get_experiences_samples(minibatch_idxs, *experiences)
                batch_log_probs = self._get_logprobs(
                    batch_ids,
                    batch_size=batch_size,
                    use_reference=False,
                    eval_mode=False,
                )
                loss, kl = self._grpo_loss(
                    batch_action_mask,
                    batch_log_probs,
                    batch_old_log_probs,
                    batch_reference_log_probs,
                    batch_advantages,
                )
                if not loss.isfinite():
                    raise ValueError(f"Loss is not finite: {loss}")
                self._backward_pass(loss)
                mean_loss += loss.item()
                mean_kl += kl.item()
        mean_loss /= len(completion_ids)
        mean_kl /= len(completion_ids)
        return mean_loss, mean_kl

    def test(
        self,
        env: ReasoningGym,
        loop: int = 1,
    ) -> torch.Tensor:
        """Returns fitness (test) score tensor of llm on test sub-set.

        :param env: The environment to be tested in
        :type env: ReasoningGym environment
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        :return: Mean test score of the agent
        :rtype: float
        """
        with env.eval_mode(), torch.no_grad():
            prompts = env.reset()
            rewards = []
            for _ in range(loop):
                completion_ids, _ = self.get_action(prompts, training=False)
                next_prompts, reward = env.step(completion_ids)
                prompts = next_prompts
                rewards.append(reward)
        reward_tensor = torch.cat(rewards)
        mean_fit = torch.mean(reward_tensor).item()
        self.fitness.append(mean_fit)
        return reward_tensor

    def _calculate_advantage(
        self, rewards: torch.Tensor, eps: float = 1e-8
    ) -> torch.Tensor:
        """Calculate the group relative advantage for each groups reward.

        :param rewards: Tensor of rewards.
        :type rewards: torch.Tensor
        :param eps: Epsilon to prevent zero division error, defaults to 1e-8
        :type eps: float, optional
        :return: Tensor of group relative advantages.
        :rtype: torch.Tensor
        """
        if len(rewards.shape) == 1:
            rewards = rewards.unsqueeze(0)

        advantage = (rewards - rewards.mean(dim=1).unsqueeze(1)) / (
            rewards.std(dim=1).unsqueeze(1) + eps
        )
        return advantage.flatten().unsqueeze(1)

    def _calculate_kl_divergence(
        self, log_probs: torch.Tensor, reference_log_probs: torch.Tensor
    ) -> torch.Tensor:
        """Calculate the KL divergence between the current and reference log probabilities.

        :param log_probs: Current policy log probabilities.
        :type log_probs: torch.Tensor
        :param reference_log_probs: Reference policy log probabilities.
        :type reference_log_probs: torch.Tensor
        :return: Kl divergence between the current and reference log probabilities.
        :rtype: torch.Tensor
        """
        return (
            torch.exp(reference_log_probs - log_probs)
            - (reference_log_probs - log_probs)
            - 1
        )

    def _grpo_loss(
        self,
        mask: torch.Tensor,
        log_probs: torch.Tensor,
        old_log_probs: torch.Tensor,
        reference_log_probs: torch.Tensor,
        advantages: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Calculate the GRPO loss.

        :param mask: Attention mask.
        :type mask: torch.Tensor
        :param log_probs: Log probabilities of the current policy.
        :type log_probs: torch.Tensor
        :param old_log_probs: Log probabilities of the old policy.
        :type old_log_probs: torch.Tensor
        :param reference_log_probs: Log probabilities of the reference policy.
        :type reference_log_probs: torch.Tensor
        :param advantages: Advantages.
        :type advantages: torch.Tensor
        :return: Mean loss and mean KL divergence.
        :rtype: tuple[torch.Tensor, torch.Tensor]
        """
        kl = self._calculate_kl_divergence(log_probs, reference_log_probs)
        log_probs_ratio = torch.exp(log_probs - old_log_probs)
        clipped_log_probs_ratio = log_probs_ratio.clamp(
            1 - self.clip_coef, 1 + self.clip_coef
        )
        surrogate = log_probs_ratio * advantages
        clipped_surrogate = clipped_log_probs_ratio * advantages
        loss = -torch.min(surrogate, clipped_surrogate) + self.beta * kl
        denominator = mask.sum(dim=-1)
        denominator = torch.where(
            denominator > 0, denominator, torch.ones_like(denominator)
        )
        loss = (loss * mask).sum(dim=-1) / denominator
        log_probs_ratio, clipped_log_probs_ratio, surrogate, clipped_surrogate = (
            None,
            None,
            None,
            None,
        )
        return loss.mean(), kl.mean()



import gc
import socket

import numpy as np
import pytest
import torch
import torch.nn as nn
from gymnasium import spaces

from agilerl.algorithms.core.registry import HyperparameterConfig, RLParameter
from tests.helper_functions import (
    gen_multi_agent_dict_or_tuple_spaces,
    generate_dict_or_tuple_space,
    generate_discrete_space,
    generate_multi_agent_box_spaces,
    generate_multi_agent_discrete_spaces,
    generate_multi_agent_multidiscrete_spaces,
    generate_multidiscrete_space,
    generate_random_box_space,
)


# Only clear CUDA cache when actually needed
@pytest.fixture(autouse=True, scope="function")
def cleanup():
    yield

    # Only clear CUDA cache if CUDA was actually used
    if torch.cuda.is_available() and torch.cuda.is_initialized():
        torch.cuda.empty_cache()

    # Only collect garbage periodically, not after every test
    if hasattr(cleanup, "call_count"):
        cleanup.call_count += 1
    else:
        cleanup.call_count = 1

    # Only run garbage collection every 10 tests
    if cleanup.call_count % 5 == 0:
        gc.collect()


# Shared device fixture to avoid repeated device checks
@pytest.fixture(scope="session")
def device():
    return "cuda" if torch.cuda.is_available() else "cpu"


# Common observation spaces (session-scoped for reuse)
@pytest.fixture(scope="session")
def vector_space():
    return generate_random_box_space(shape=(4,))


@pytest.fixture(scope="session")
def discrete_space():
    return generate_discrete_space(2)


@pytest.fixture(scope="session")
def dict_space():
    return generate_dict_or_tuple_space(2, 2, dict_space=True)


@pytest.fixture(scope="session")
def tuple_space():
    return generate_dict_or_tuple_space(2, 2, dict_space=False)


@pytest.fixture(scope="session")
def multidiscrete_space():
    return generate_multidiscrete_space(2, 2)


@pytest.fixture(scope="session")
def multibinary_space():
    return spaces.MultiBinary(4)


@pytest.fixture(scope="session")
def image_space():
    return generate_random_box_space(shape=(3, 32, 32), low=0, high=255)


# Common multi-agent spaces
@pytest.fixture(scope="session")
def ma_vector_space():
    return generate_multi_agent_box_spaces(3, (6,))


@pytest.fixture(scope="session")
def ma_discrete_space():
    return generate_multi_agent_discrete_spaces(3, 2)


@pytest.fixture(scope="session")
def ma_multidiscrete_space():
    return generate_multi_agent_multidiscrete_spaces(3, 2)


@pytest.fixture(scope="session")
def ma_multibinary_space():
    return [spaces.MultiBinary(2) for _ in range(3)]


@pytest.fixture(scope="session")
def ma_image_space():
    return generate_multi_agent_box_spaces(3, (3, 32, 32), low=0, high=255)


@pytest.fixture(scope="session")
def ma_dict_space():
    return gen_multi_agent_dict_or_tuple_spaces(3, 2, 2, dict_space=True)


@pytest.fixture(scope="session")
def ma_dict_space_small():
    return gen_multi_agent_dict_or_tuple_spaces(3, 1, 1, dict_space=True)


# Simple network fixtures (function-scoped to avoid state issues)
@pytest.fixture(scope="function")
def simple_mlp():
    return nn.Sequential(
        nn.Linear(4, 20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.ReLU(),
        nn.Linear(10, 2),
        nn.Softmax(dim=-1),
    )


@pytest.fixture(scope="function")
def simple_mlp_critic():
    return nn.Sequential(
        nn.Linear(6, 20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.ReLU(),
        nn.Linear(10, 1),
        nn.Tanh(),
    )


@pytest.fixture(scope="function")
def simple_cnn():
    return nn.Sequential(
        nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1),
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Conv2d(16, 32, kernel_size=3, stride=1, padding=1),
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Flatten(),
        nn.Linear(32 * 8 * 8, 128),
        nn.ReLU(),
        nn.Linear(128, 2),
        nn.Softmax(dim=-1),
    )


########################################################
################# MUTATIONS ############################
########################################################


@pytest.fixture(scope="session")
def ac_hp_config():
    return HyperparameterConfig(
        lr_actor=RLParameter(min=1e-4, max=1e-2),
        lr_critic=RLParameter(min=1e-4, max=1e-2),
        batch_size=RLParameter(min=8, max=512, dtype=int),
        learn_step=RLParameter(
            min=20, max=200, dtype=int, grow_factor=1.5, shrink_factor=0.75
        ),
    )


@pytest.fixture(scope="session")
def default_hp_config():
    yield HyperparameterConfig(
        lr=RLParameter(min=6.25e-5, max=1e-2),
        batch_size=RLParameter(min=8, max=512, dtype=int),
        learn_step=RLParameter(
            min=1, max=10, dtype=int, grow_factor=1.5, shrink_factor=0.75
        ),
    )


@pytest.fixture(scope="session")
def grpo_hp_config():
    yield HyperparameterConfig(
        lr=RLParameter(min=0.00001, max=1),
    )


@pytest.fixture(scope="session")
def encoder_mlp_config():
    yield {
        "latent_dim": 8,
        "min_latent_dim": 1,
        "encoder_config": {"hidden_size": [8, 8], "min_mlp_nodes": 1},
        "head_config": {"hidden_size": [8, 8], "min_mlp_nodes": 1},
    }


@pytest.fixture(scope="session")
def encoder_simba_config():
    yield {
        "latent_dim": 8,
        "min_latent_dim": 1,
        "simba": True,
        "encoder_config": {
            "hidden_size": 64,
            "num_blocks": 3,
        },
        "head_config": {"hidden_size": [8], "min_mlp_nodes": 1},
    }


@pytest.fixture(scope="session")
def encoder_cnn_config():
    yield {
        "latent_dim": 8,
        "min_latent_dim": 1,
        "encoder_config": {
            "channel_size": [5, 5],
            "kernel_size": [3, 3],
            "stride_size": [1, 1],
            "min_channel_size": 1,
            "max_channel_size": 10,
        },
        "head_config": {"hidden_size": [8, 8], "min_mlp_nodes": 1},
    }


@pytest.fixture(scope="session")
def encoder_multi_input_config():
    yield {
        "latent_dim": 8,
        "min_latent_dim": 1,
        "encoder_config": {
            "cnn_config": {
                "channel_size": [5],
                "kernel_size": [3],
                "stride_size": [1],
                "min_channel_size": 1,
                "max_channel_size": 10,
            },
            "mlp_config": {"hidden_size": [8, 8], "min_mlp_nodes": 1},
        },
        "head_config": {"hidden_size": [8, 8], "min_mlp_nodes": 1},
    }


class EvoDummyRNG:
    rng = np.random.default_rng(seed=42)

    def choice(self, a, size=None, replace=True, p=None):
        return 1

    def integers(self, low=0, high=None):
        return self.rng.integers(low, high)


@pytest.fixture(scope="session")
def dummy_rng():
    return EvoDummyRNG()


dist_env = dict(
    ACCELERATE_USE_DEEPSPEED="true",
    MASTER_ADDR="localhost",
    MASTER_PORT="10999",
    RANK="0",
    LOCAL_RANK="0",
    WORLD_SIZE="1",
    PYTORCH_CUDA_ALLOC_CONF="expandable_segments:True",
    CUDA_VISIBLE_DEVICES="0",
)


def get_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture
def deepspeed_env():
    import os

    dynamic_dist_env = dist_env.copy()
    dynamic_dist_env["MASTER_PORT"] = str(get_free_port())
    existing_vars = {}
    for key, value in dynamic_dist_env.items():
        key = key.upper()
        if key in os.environ:
            existing_vars[key] = os.environ[key]
        os.environ[key] = str(value)

    try:
        yield
    finally:
        for key in dynamic_dist_env:
            key = key.upper()
            if key in existing_vars:
                # restore previous value
                os.environ[key] = existing_vars[key]
            else:
                os.environ.pop(key, None)
